<?php

/* +***********************************************************************************
 * The contents of this file are subject to the vtiger CRM Public License Version 1.0
 * ("License"); You may not use this file except in compliance with the License
 * The Original Code is:  vtiger CRM Open Source
 * The Initial Developer of the Original Code is vtiger.
 * Portions created by vtiger are Copyright (C) vtiger.
 * All Rights Reserved.
 * *********************************************************************************** */

/**
 * Provides API to work with HTTP Connection.
 * @package vtlib
 */
class Vtiger_Net_Client {

    var $client;
    var $url;
    var $response;

    /**
     * Constructor
     * @param String URL of the site
     * Example: 
     * $client = new Vtiger_New_Client('http://www.vtiger.com');
     */
    function __construct($url) {
        $this->setURL($url);
    }

    /**
     * Set another url for this instance
     * @param String URL to use go forward
     */
    function setURL($url) {
        $this->url = $url;
        $this->client = new HTTP_Request2('', '', array('follow_redirects' => true));
        $this->client->setURL($url);
        $this->response = false;
        $this->setDefaultHeaders();
    }

    function setDefaultHeaders() {
        $headers = array();
        if (isset($_SERVER)) {
            global $site_URL;
            $headers['referer'] = isset($_SERVER['HTTP_REFERER']) ? $_SERVER['HTTP_REFERER'] : ($site_URL . "?noreferer");
            if (isset($_SERVER['HTTP_USER_AGENT'])) {
                $headers['user-agent'] = $_SERVER['HTTP_USER_AGENT'];
            }
        } else {
            global $site_URL;
            $headers['referer'] = ($site_URL . "?noreferer");
        }
        $this->setHeaders($headers);
    }

    /**
     * Function to send http request and handling errors.
     * Error Handling : Request2 will throw HTTP_Request2_Exception instead of PEAR Errors. We are catching those exceptions and setting it to response message.
     */
    public function send() {
        try {
            $this->response = $this->client->send();
        } catch (HTTP_Request2_Exception $ex) {
            $this->response = new Vtiger_Net_ClientException($ex);
        } catch (Exception $ex) {
            $this->response = new Vtiger_Net_ClientException($ex);
        }
    }

    /**
     * Set custom HTTP Headers
     * @param Map HTTP Header and Value Pairs
     */
    function setHeaders($values) {
        foreach ($values as $key => $value) {
            $this->client->setHeader($key, $value);
        }
    }

    /**
     * Perform a GET request
     * @param Map key-value pair or false
     * @param Integer timeout value
     */
    function doGet($params = array(), $timeout = null) {
        if ($timeout) {
            $this->client->setConfig('connect_timeout', $timeout);
        }
        $this->client->setURL($this->url);
        $this->client->setMethod(HTTP_Request2::METHOD_GET);
        if ($params) {
            foreach ($params as $key => $value) {
                $this->client->getUrl()->setQueryVariable($key, $value);
            }
        }
        $this->send();
        $content = $this->wasError() ? false : $this->response->getBody();
        $this->disconnect();
        return $content;
    }

    /**
     * Perform a POST request
     * @param Map key-value pair or false
     * @param Integer timeout value
     */
    function doPost($params = array(), $timeout = null) {
        if ($timeout) {
            $this->client->setConfig('connect_timeout', $timeout);
        }
        $this->client->setURL($this->url);
        $this->client->setMethod(HTTP_Request2::METHOD_POST);
        if ($params) {
            if (is_string($params)) {
                $this->client->setBody($params);
            } else {
                foreach ($params as $key => $value) {
                    $this->client->addPostParameter($key, $value);
                }
            }
        }
        $this->send();
        $content = $this->wasError() ? false : $this->response->getBody();
        $this->disconnect();
        return $content;
    }

    /**
     * Perform a PUT request
     * @param Map key-value pair or false
     * @param Integer timeout value
     */
    function doPut($params = array(), $timeout = null) {
        if ($timeout) {
            $this->client->setConfig('connect_timeout', $timeout);
        }
        $this->client->setURL($this->url);
        $this->client->setMethod(HTTP_Request2::METHOD_PUT);

        if ($params) {
            if (is_string($params)) {
                $this->client->setBody($params);
            } else {
                foreach ($params as $key => $value) {
                    $this->client->addPostParameter($key, $value);
                }
            }
        }
        $this->send();
        $content = $this->wasError() ? false : $this->response->getBody();
        $this->disconnect();
        return $content;
    }

    /**
     * Perform a Patch request
     * @param Map key-value pair or false
     * @param Integer timeout value
     */
    function doPatch($params = array(), $timeout = null) {
        if ($timeout) {
            $this->client->setConfig('connect_timeout', $timeout);
        }
        $this->client->setURL($this->url);
        $this->client->setMethod("PATCH");
        if ($params) {
            if (is_string($params)) {
                $this->client->setBody($params);
            } else {
                foreach ($params as $key => $value) {
                    $this->client->addPostParameter($key, $value);
                }
            }
        }
        $this->send();
        $content = $this->wasError() ? false : $this->response->getBody();
        $this->disconnect();
        return $content;
    }

    /**
     * Perform a POST request
     * @param Map key-value pair or false
     * @param Integer timeout value
     */
    function doDelete($params = array(), $timeout = null) {
        if ($timeout) {
            $this->client->setConfig('connect_timeout', $timeout);
        }
        $this->client->setURL($this->url);
        $this->client->setMethod(HTTP_Request2::METHOD_DELETE);
        if ($params) {
            foreach ($params as $key => $value) {
                $this->client->addPostParameter($key, $value);
            }
        }
        $this->send();
        $content = $this->wasError() ? false : $this->response->getBody();
        $this->disconnect();
        return $content;
    }

    /**
     * Add File to Send with a post
     * @param String file upload fieldname
     * @param Mixed path of file to add
     * @param Mixed file content type of file being uploaded(default : application/octet-stream)
     */
    function addFiles($inputName, $filePath, $fileContentType = 'application/octet-stream') {
        $this->client->addUpload($inputName, $filePath, '', $fileContentType);
        $this->client->setHeader('Content-Type', 'multipart/form-data');
    }

    /**
     * Did last request resulted in error?
     */
    function wasError() {
        return isset($this->response->message) ? true : false;
    }

    /**
     * Disconnect this instance
     */
    function disconnect() {
        $this->client->setLastEvent('disconnect');
    }

    /**
     * Returns the response code
     * @access public
     * @return mixed     Response code, false if not set
     */
    function getResponseCode() {
        $responseCode = $this->response->getStatus();
        return isset($responseCode) ? $responseCode : false;
    }

    /**
     * Function used to set basic auth
     * @param type $userName
     * @param type $password
     */
    function setBasicAuth($userName, $password) {
        $this->client->setAuth($userName, $password);
    }

    /**
     * Function used to log the build request.
     * @return type <Array>
     */
    function buildRequest() {
        $request = $this->client->getMethod() . " " . $this->client->getUrl()->getPath() . "\n";
        $request .= "Host: " . $this->client->getUrl()->getHost() . "\n";
        $headers = $this->client->getHeaders();
        foreach ($headers as $key => $value) {
            $request .= $key . ": " . $value . "\n";
        }
        $body = $this->client->getBody();
        if (is_string($body)) {
            $request .= "\n" . $body;
        } else if ($body && is_object($body) && get_class($body) === 'HTTP_Request2_MultipartBody') {
            $request .= "\n" . $body->read(1000);
        }
        return $request;
    }

    /**
     * Function to get response header
     * @param type $name
     * @return type
     */
    public function getResponseHeader($name) {
        return $this->response->getHeader($name);
    }

    /**
     * Functions to get Request headers
     */
    public function getHeaders() {
        return $this->client->getHeaders();
    }
}

class Vtiger_Net_ClientException {

    var $code;
    var $message;

    public function __construct($ex) {
        $this->code = $ex->getCode();
        $this->message = $ex->getMessage();
    }

    /**
     * Function to return response code.
     * @return type
     */
    public function getResponseCode() {
        return $this->code;
    }

    /**
     * Function to return response message.
     * @return type
     */
    public function getMessage() {
        return $this->message;
    }

    /**
     * Function to return reponse header.
     * @return type
     */
    public function getHeader($name) {
        return '';
    }
        
    /**
     * Function to return response code
     */
    public function getStatus() {
        return $this->code;
    }
}
